# -*- coding: utf-8 -*-

from .serializer import Serializer

from .dom import addClass
from .dom import addOldIDs
from .dom import appendChild
from .dom import appendContents
from .dom import approximateLineNumber
from .dom import childElements
from .dom import childNodes
from .dom import circledDigits
from .dom import clearContents
from .dom import closestAncestor
from .dom import closestAttr
from .dom import createElement
from .dom import dedupIDs
from .dom import E
from .dom import emptyText
from .dom import escapeAttr
from .dom import escapeCSSIdent
from .dom import escapeHTML
from .dom import escapeUrlFrag
from .dom import filterAncestors
from .dom import find
from .dom import findAll
from .dom import fixSurroundingTypography
from .dom import fixTypography
from .dom import fixupIDs
from .dom import foldWhitespace
from .dom import hasAncestor
from .dom import hasAttrs
from .dom import hasChildElements
from .dom import hasClass
from .dom import hashContents
from .dom import hasOnlyChild
from .dom import headingLevelOfElement
from .dom import innerHTML
from .dom import insertAfter
from .dom import insertBefore
from .dom import isElement
from .dom import isEmpty
from .dom import isNormative
from .dom import isOddNode
from .dom import moveContents
from .dom import nodeIter
from .dom import outerHTML
from .dom import parentElement
from .dom import parseDocument
from .dom import parseHTML
from .dom import prependChild
from .dom import previousElements
from .dom import relevantHeadings
from .dom import removeAttr
from .dom import removeClass
from .dom import removeNode
from .dom import replaceAwkwardCSSShorthands
from .dom import replaceContents
from .dom import replaceMacros
from .dom import replaceNode
from .dom import safeID
from .dom import scopingElements
from .dom import sectionName
from .dom import serializeTag
from .dom import textContent
from .dom import textContentIgnoringDecorative
from .dom import treeAttr
from .dom import unescape
from .dom import unfixTypography
from .dom import wrapContents