# -*- coding: utf-8 -*-


from .main import fixupDataFiles, update, updateReadonlyDataFiles
from .manifest import createManifest

__all__ = ["updateBackRefs", "updateCrossRefs", "updateBiblio", "updateCanIUse", "updateLinkDefaults", "updateTestSuites", "updateLanguages"]
